<?php

namespace Drupal\janrain_connect_ui\EventSubscriber;

use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\user\UserDataInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpFoundation\Session\Session;
use Drupal\janrain_connect\Constants\JanrainConnectConstants;
use Drupal\janrain_connect\Constants\JanrainConnectWebServiceConstants;
use Drupal\janrain_connect\Service\JanrainConnectConnector;
use Drupal\janrain_connect\Service\JanrainConnectLogin;
use Drupal\janrain_connect_ui\Event\JanrainConnectUiEvents;
use Drupal\janrain_connect_ui\Event\JanrainConnectUiSubmitEvent;
use Drupal\janrain_connect_ui\Service\JanrainConnectUiUsers;
use Drupal\Core\Url;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\Yaml\Yaml;
use Drupal\janrain_connect_ui\Service\JanrainConnectUiTokenService;
use Drupal\janrain_connect_ui\Exception\AuthenticationVerifyAccountException;
use Drupal\janrain_connect_ui\Exception\AuthenticationDrupalLoginException;

/**
 * Subscribe to Janrain Connect Submit.
 */
class JanrainConnectUiSubscriber implements EventSubscriberInterface {

  use StringTranslationTrait;

  /**
   * The user data service.
   *
   * @var \Drupal\user\UserDataInterface
   */
  protected $userData;

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * JanrainConnectConnector.
   *
   * @var \Drupal\janrain_connect\Service\JanrainConnectConnector
   */
  private $janrainConnector;

  /**
   * JanrainConnectUsers.
   *
   * @var \Drupal\janrain_connect_ui\Service\JanrainConnectUiUsers
   */
  private $janrainUsers;

  /**
   * Module Handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  private $moduleHandler;

  /**
   * Symfony session handler.
   *
   * @var \Symfony\Component\HttpFoundation\Session\Session
   */
  private $session;

  /**
   * JanrainConnectLogin.
   *
   * @var \Drupal\janrain_connect\Service\JanrainConnectLogin
   */
  private $janrainLogin;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $config;

  /**
   * JanrainConnectToken.
   *
   * @var \Drupal\janrain_connect_ui\Service\JanrainConnectUiTokenService
   */
  protected $janrainConnectUiTokenService;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    JanrainConnectConnector $janrain_connector,
    Session $session,
    ModuleHandlerInterface $module_handler,
    JanrainConnectUiUsers $janrain_users,
    RequestStack $request_stack,
    UserDataInterface $user_data,
    JanrainConnectLogin $janrain_login,
    ConfigFactory $config_factory,
    JanrainConnectUiTokenService $janrain_connect_ui_token_service
  ) {
    $this->janrainConnector = $janrain_connector;
    $this->session = $session;
    $this->moduleHandler = $module_handler;
    $this->janrainUsers = $janrain_users;
    $this->requestStack = $request_stack;
    $this->userData = $user_data;
    $this->janrainLogin = $janrain_login;
    $this->config = $config_factory->get('janrain_connect.settings');
    $this->janrainConnectUiTokenService = $janrain_connect_ui_token_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {

    $events[JanrainConnectUiEvents::EVENT_SUBMIT][] = ['submitRequestToJanrain'];

    return $events;
  }

  /**
   * Submit Registration.
   */
  public function submitRequestToJanrain(JanrainConnectUiSubmitEvent $event) {
    $data = $event->getData();
    $form = $event->getForm();
    $form_state = $event->getFormState();

    switch ($event->getFormId()) {
      case JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_REGISTRATION:
        $this->sendRegistrationToJanrain($data, $form, $form_state);
        break;

      case JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_SIGNIN:
        $this->sendLoginToJanrain($data, $form, $form_state);
        break;

      case JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_EDIT_PROFILE:
        $this->sendProfileUpdateToJanrain($data, $form, $form_state);
        break;

      case JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_FORGOT_PASSWORD:
        $this->sendForgotPasswordToJanrain($data, $form, $form_state);
        break;

      case JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_CHANGE_PASSWORD:
        $this->sendChangePasswordToJanrain($data, $form, $form_state);
        break;
    }
  }

  /**
   * Send registration data to Janrain.
   *
   * @param array $data
   *   User data to create a janrain registration.
   * @param mixed $form
   *   Form.
   * @param mixed $form_state
   *   Form State.
   */
  private function sendRegistrationToJanrain(array $data, $form, $form_state) {
    // Try register in Janrain.
    $result = $this->janrainConnector->register($data);

    if (!empty($result['has_errors'])) {
      $this->printErrorMessages(
        $result,
        $form_state,
        JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_REGISTRATION
      );
      return;
    }

    // Persist user in Drupal site.
    $status = $this->janrainUsers->persistUser($data, $result, FALSE);

    if ($status !== SAVED_NEW) {
      $form_state->setErrorByName(
        JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_REGISTRATION,
        $this->t(JanrainConnectConstants::JANRAIN_CONNECT_REGISTRATION_GENERIC_ERROR)
      );
      return;
    }

    try {
      $user = user_load_by_name($result['capture_user']->uuid);
      $this->verifyConstraintsAndDrupalLogin($result, $user);
      $success_message = $this->getSuccessMessage($form_state, JanrainConnectConstants::JANRAIN_CONNECT_REGISTRATION_MESSAGE_SUCCESS);
      $success_message .= '<script>jQuery(window).bind("dialog:beforeclose", function(context) {location.reload();});</script>';
      drupal_set_message($success_message, 'status');
    }
    catch (AuthenticationVerifyAccountException $e) {
      $verify_email_message = $error = $this->t(
        JanrainConnectConstants::JANRAIN_CONNECT_AUTHENTICATION_VERIFY_ACCOUNT_ERROR
      );
      $form_id = $form_state->get('form_id');
      $ajax_settings = $this->getJanrainConnectUiAjaxSettings($form_id);
      if (!empty($ajax_settings['verify_email_message'])) {
        $tokens = $this->janrainConnectUiTokenService->getMessageFormTokens($ajax_settings['verify_email_message'], $form_state);
        $verify_email_message = $this->t($ajax_settings['verify_email_message'], $tokens);
      }
      drupal_set_message($verify_email_message, 'status');
    }
    catch (AuthenticationDrupalLoginException $e) {
      $form_state->setErrorByName(
        JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_REGISTRATION,
        $this->t(
          JanrainConnectConstants::JANRAIN_CONNECT_REGISTRATION_GENERIC_ERROR
        )
      );
    }
  }

  /**
   * Send login data to Janrain.
   *
   * @param array $data
   *   Required data to allow user login.
   * @param mixed $form
   *   Form.
   * @param mixed $form_state
   *   Form State.
   */
  private function sendLoginToJanrain(array $data, $form, $form_state) {
    // Try login in Janrain.
    $result = $this->janrainConnector->login($data);

    if (!empty($result['has_errors'])) {
      $this->printErrorMessages(
        $result,
        $form_state,
        JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_SIGNIN
      );
      return;
    }

    // Load by e-mail because we don't have the UUID.
    $user = user_load_by_mail($data['signInEmailAddress']);

    // If user exists on janrain but not exists in Drupal, create user.
    if (empty($user)) {
      $data_register['emailAddress'] = $data['signInEmailAddress'];
      $user = $this->janrainUsers->persistUser($data_register, $result, TRUE);
    }

    if (empty($user)) {
      $form_state->setErrorByName(
        JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_SIGNIN,
        $this->t(JanrainConnectConstants::JANRAIN_CONNECT_AUTHENTICATION_GENERIC_ERROR)
      );
      return;
    }

    $error = NULL;
    try {
      $this->verifyConstraintsAndDrupalLogin($result, $user);

      $success_message = $this->getSuccessMessage(
        $form_state,
        'You are successfully logged in'
      );
      drupal_set_message($success_message, 'status');
    }
    catch (AuthenticationVerifyAccountException $e) {
      $error = $this->t(
        JanrainConnectConstants::JANRAIN_CONNECT_AUTHENTICATION_VERIFY_ACCOUNT_ERROR
      );
    }
    catch (AuthenticationDrupalLoginException $e) {
      $error = $this->t(
        JanrainConnectConstants::JANRAIN_CONNECT_AUTHENTICATION_GENERIC_ERROR
      );
    }

    if (!empty($error)) {
      $form_state->setErrorByName(
        JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_SIGNIN,
        $error
      );
      return;
    }

    // If are using ajax, the return is by forms config yml.
    if ($this->formIsUsingAjax($form_state)) {
      return;
    }

    $url = Url::fromRoute('<front>', [], []);
    $response = new RedirectResponse($url->toString());
    $response->send();
    $this->requestStack->getCurrentRequest()->query->set('destination', '/');
  }

  /**
   * Verify form is using ajax.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form State.
   *
   * @return bool
   *   Form is using ajax.
   */
  private function formIsUsingAjax(FormStateInterface $form_state) {
    $ajax_settings = $this
      ->getJanrainConnectUiAjaxSettings($form_state->get('form_id'));

    return !empty($ajax_settings['use_ajax']);
  }

  /**
   * Verify constraints and Drupal login.
   *
   * @param array $result
   *   Result returned by Janrain API.
   * @param object $user
   *   Drupal user.
   *
   * @throws \Drupal\janrain_connect_ui\Exception\AuthenticationVerifyAccountException
   * @throws \Drupal\janrain_connect_ui\Exception\AuthenticationDrupalLoginException
   */
  private function verifyConstraintsAndDrupalLogin(array $result, $user) {
    if (!$this->emailVerified($result['capture_user'])) {
      throw new AuthenticationVerifyAccountException();
    }

    if (!$this->janrainLogin->login($result['access_token'], $user)) {
      throw new AuthenticationDrupalLoginException();
    }
  }

  /**
   * Check if email verified.
   *
   * @param object $capture_user
   *   User returned by Janrain API.
   *
   * @return bool
   *   If email verified is disable is TRUE or email was verified in Janrain.
   */
  private function emailVerified($capture_user) {
    $check_email_verified = $this->config->get('config_auth_check_email_verified');
    return empty($check_email_verified) || !empty($capture_user->emailVerified);
  }

  /**
   * Get Janrain Ui Ajax Settings.
   *
   * @param string $form_id
   *   Form ID.
   *
   * @return mixed
   *   Return value.
   */
  private function getJanrainConnectUiAjaxSettings(string $form_id) {
    $configuration_forms = $this->config->get('configuration_forms');
    $ajax_settings = Yaml::parse($configuration_forms);

    if (!empty($ajax_settings[$form_id])) {
      return $ajax_settings[$form_id];
    }

    return '';
  }

  /**
   * Send profile update to Janrain.
   *
   * @param array $data
   *   User data to update a profile in Janrain.
   * @param mixed $form
   *   Form.
   * @param mixed $form_state
   *   Form State.
   */
  private function sendProfileUpdateToJanrain(array $data, $form, $form_state) {
    $accessToken = $this->session->get('janrain_connect_access_token');

    // TODO: Remove this condition after the whole login process is done.
    if (empty($accessToken)) {
      // @codingStandardsIgnoreLine.
      drupal_set_message('You are not logged in', 'warning');
      return;
    }

    $result = $this->janrainConnector->updateUserProfile($accessToken, $data);

    if (!$result['has_errors']) {
      // Message success should be translatable @codingStandardsIgnoreLine.
      $success_message = $this->getSuccessMessage($form_state, 'Profile successfully updated');
      drupal_set_message($success_message, 'status');

      // Todo: Implements Dependecy Injection. @codingStandardsIgnoreLine
      $url = Url::fromUri('base:' . \Drupal::service('path.current')->getPath());
      $response = new RedirectResponse($url->toString());
      $response->send();
    }
    else {
      $this->printErrorMessages(
        $result,
        $form_state,
        JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_EDIT_PROFILE
      );
    }
  }

  /**
   * Send Forgot Password to Janrain.
   *
   * @param array $data
   *   User data to create a janrain registration.
   */
  // @codingStandardsIgnoreLine
  private function sendForgotPasswordToJanrain(array $data, $form, $form_state) {

    $result = $this->janrainConnector->forgotPassword($data['signInEmailAddress']);

    if (!$result['has_errors']) {
      $success_message = $this->getSuccessMessage($form_state, "We've sent an email with instructions to create a new password. Your existing password has not been changed.");
      drupal_set_message($success_message, 'status');
    }
    else {
      $this->printErrorMessages(
        $result,
        $form_state,
        JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_FORGOT_PASSWORD
      );
    }
  }

  /**
   * Send Change Password to Janrain.
   *
   * @param array $data
   *   User data to change password.
   * @param mixed $form
   *   Form.
   * @param mixed $form_state
   *   Form State.
   */
  private function sendChangePasswordToJanrain(array $data, $form, $form_state) {

    $actual_password = $data['currentPassword'];
    $new_password = $data['newPassword'];
    $new_password_confirm = $data['newPasswordConfirm'];
    $token = $this->session->get('janrain_connect_access_token');

    $result = $this->janrainConnector->changePassword($actual_password, $new_password, $new_password_confirm, $token);

    if (!$result['has_errors']) {
      $success_message = $this->getSuccessMessage($form_state, JanrainConnectConstants::JANRAIN_CONNECT_CHANGE_PASSWORD_MESSAGE_SUCCESS);
      drupal_set_message($success_message, 'status');
    }
    else {
      $this->printErrorMessages(
        $result,
        $form_state,
        JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_CHANGE_PASSWORD
      );
    }
  }

  /**
   * Print error messages.
   *
   * @param array $result
   *   Result returned by Janrain API.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form State.
   * @param string $form_id
   *   The form id.
   */
  private function printErrorMessages(array $result, FormStateInterface $form_state, $form_id) {

    // If exists janrain_connect validate, use it to get messages.
    if ($this->moduleHandler->moduleExists('janrain_connect_validate')) {

      // Use direct access because is possible uninstall the Janrain Connect
      // Validate. @codingStandardsIgnoreLine
      $messages = \Drupal::service('janrain_connect_validate.messages_mapping')->getMessagesFields($result, JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_REGISTRATION);
    }
    else {
      $messages[$form_id] = $this->getGenericErrorMessage($form_id);
    }

    $configuration_form_field = FALSE;

    if (!empty($form_id)) {

      $configuration_fields = $this->config->get('configuration_fields');

      $configuration_fields = Yaml::parse($configuration_fields);

      $configuration_form_field = $configuration_fields[$form_id];

    }

    foreach ($messages as $field_id => $field_messages) {
      foreach ($field_messages as $message) {
        if (!empty($configuration_form_field[$field_id]['validation-update-message'][$message])) {
          $message = $configuration_form_field[$field_id]['validation-update-message'][$message];
        }

        if (!empty($field_id) && !empty($form_id) && $field_id == $form_id) {

          $configuration_forms = $this->config->get('configuration_forms');
          $configuration_forms = Yaml::parse($configuration_forms);

          if (!empty($configuration_forms[$form_id])) {
            $configuration_form = $configuration_forms[$form_id];
            if (!empty($configuration_form['default_field_error'])) {
              $field_id = $configuration_form['default_field_error']; 
            }
          }
        }

        // @codingStandardsIgnoreLine.
        $form_state->setErrorByName($field_id, $this->t($message));
      }
    }

  }

  /**
   * Get generic error message by form id.
   *
   * @param string $form_id
   *   Form id.
   *
   * @return string
   *   Generic error message.
   */
  private function getGenericErrorMessage($form_id) {
    $generic_error_messages = [
      JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_REGISTRATION => JanrainConnectConstants::JANRAIN_CONNECT_REGISTRATION_GENERIC_ERROR,
      JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_SIGNIN => JanrainConnectConstants::JANRAIN_CONNECT_AUTHENTICATION_GENERIC_ERROR,
      JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_EDIT_PROFILE => JanrainConnectConstants::JANRAIN_CONNECT_EDIT_PROFILE_GENERIC_ERROR,
      JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_FORGOT_PASSWORD => JanrainConnectConstants::JANRAIN_CONNECT_FORGOT_PASSWORD_GENERIC_ERROR,
      JanrainConnectWebServiceConstants::JANRAIN_CONNECT_FORM_CHANGE_PASSWORD => JanrainConnectConstants::JANRAIN_CONNECT_CHANGE_PASSWORD_GENERIC_ERROR,
    ];

    return (isset($generic_error_messages[$form_id]) ?
      $generic_error_messages[$form_id] : '');
  }

  /**
   * Get success message.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form State.
   * @param string $default_message
   *   The default message.
   *
   * @return string
   *   Return success message from ajax settings or default.
   */
  private function getSuccessMessage(FormStateInterface $form_state, $default_message) {
    $form_id = $form_state->get('form_id');
    $ajax_settings = $this->getJanrainConnectUiAjaxSettings($form_id);

    $success_message = $this->t($default_message);
    if (!empty($ajax_settings['success_message'])) {
      $tokens = $this->janrainConnectUiTokenService->getMessageFormTokens($ajax_settings['success_message'], $form_state);
      $success_message = $this->t($ajax_settings['success_message'], $tokens);
    }

    return $success_message;
  }

}
